@@ -19,23 +19,19 @@
 #include "Database.h"
 
 #include "core/AsyncTask.h"
-#include "core/Clock.h"
 #include "core/FileWatcher.h"
 #include "core/Group.h"
-#include "core/Merger.h"
-#include "core/Metadata.h"
+#include "crypto/Random.h"
 #include "format/KdbxXmlReader.h"
 #include "format/KeePass2Reader.h"
 #include "format/KeePass2Writer.h"
-#include "keys/FileKey.h"
-#include "keys/PasswordKey.h"
 
-#include <QFile>
 #include <QFileInfo>
+#include <QJsonObject>
+#include <QRegularExpression>
 #include <QSaveFile>
 #include <QTemporaryFile>
 #include <QTimer>
-#include <QXmlStreamReader>
 
 QHash<QUuid, QPointer<Database>> Database::s_uuidMap;
 
@@ -44,24 +40,39 @@ Database::Database()
     , m_data()
     , m_rootGroup(nullptr)
     , m_fileWatcher(new FileWatcher(this))
-    , m_emitModified(false)
     , m_uuid(QUuid::createUuid())
 {
-    setRootGroup(new Group());
-    rootGroup()->setUuid(QUuid::createUuid());
-    rootGroup()->setName(tr("Passwords", "Root group name"));
+    // setup modified timer
     m_modifiedTimer.setSingleShot(true);
+    connect(this, &Database::emitModifiedChanged, this, [this](bool value) {
+        if (!value) {
+            stopModifiedTimer();
+        }
+    });
+    connect(&m_modifiedTimer, &QTimer::timeout, this, &Database::emitModified);
 
-    s_uuidMap.insert(m_uuid, this);
-
-    connect(m_metadata, SIGNAL(metadataModified()), SLOT(markAsModified()));
-    connect(&m_modifiedTimer, SIGNAL(timeout()), SIGNAL(databaseModified()));
-    connect(this, SIGNAL(databaseOpened()), SLOT(updateCommonUsernames()));
-    connect(this, SIGNAL(databaseSaved()), SLOT(updateCommonUsernames()));
+    // other signals
+    connect(m_metadata, &Metadata::modified, this, &Database::markAsModified);
+    connect(this, &Database::databaseOpened, this, [this]() {
+        updateCommonUsernames();
+        updateTagList();
+    });
+    connect(this, &Database::modified, this, [this] { updateTagList(); });
+    connect(this, &Database::databaseSaved, this, [this]() { updateCommonUsernames(); });
     connect(m_fileWatcher, &FileWatcher::fileChanged, this, &Database::databaseFileChanged);
 
+    // static uuid map
+    s_uuidMap.insert(m_uuid, this);
+
+    // block modified signal and set root group
+    setEmitModified(false);
+
+    // Note: oldGroup is nullptr but need to respect return value capture
+    auto oldGroup = setRootGroup(new Group());
+    Q_UNUSED(oldGroup)
+
     m_modified = false;
-    m_emitModified = true;
+    setEmitModified(true);
 }
 
 Database::Database(const QString& filePath)
@@ -86,17 +97,16 @@ QUuid Database::uuid() const
  * read-write mode and fall back to read-only if that is not possible.
  *
  * @param key composite key for unlocking the database
- * @param readOnly open in read-only mode
  * @param error error message in case of failure
  * @return true on success
  */
-bool Database::open(QSharedPointer<const CompositeKey> key, QString* error, bool readOnly)
+bool Database::open(QSharedPointer<const CompositeKey> key, QString* error)
 {
     Q_ASSERT(!m_data.filePath.isEmpty());
     if (m_data.filePath.isEmpty()) {
         return false;
     }
-    return open(m_data.filePath, std::move(key), error, readOnly);
+    return open(m_data.filePath, std::move(key), error);
 }
 
 /**
@@ -106,11 +116,10 @@ bool Database::open(QSharedPointer<const CompositeKey> key, QString* error, bool
  *
  * @param filePath path to the file
  * @param key composite key for unlocking the database
- * @param readOnly open in read-only mode
  * @param error error message in case of failure
  * @return true on success
  */
-bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey> key, QString* error, bool readOnly)
+bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey> key, QString* error)
 {
     QFile dbFile(filePath);
     if (!dbFile.exists()) {
@@ -144,7 +153,6 @@ bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey>
         return false;
     }
 
-    setReadOnly(readOnly);
     setFilePath(filePath);
     dbFile.close();
 
@@ -157,6 +165,27 @@ bool Database::open(const QString& filePath, QSharedPointer<const CompositeKey>
     return true;
 }
 
+/**
+ * KDBX format version.
+ */
+quint32 Database::formatVersion() const
+{
+    return m_data.formatVersion;
+}
+
+void Database::setFormatVersion(quint32 version)
+{
+    m_data.formatVersion = version;
+}
+
+/**
+ * Whether the KDBX minor version is greater than the newest supported.
+ */
+bool Database::hasMinorVersionMismatch() const
+{
+    return m_data.formatVersion > KeePass2::FILE_VERSION_MAX;
+}
+
 bool Database::isSaving()
 {
     bool locked = m_saveMutex.tryLock();
@@ -172,10 +201,10 @@ bool Database::isSaving()
  *
  * @param error error message in case of failure
  * @param atomic Use atomic file transactions
- * @param backup Backup the existing database file, if exists
+ * @param backupFilePath Absolute file path to write the backup file to. Pass an empty QString to disable backup.
  * @return true on success
  */
-bool Database::save(QString* error, bool atomic, bool backup)
+bool Database::save(SaveAction action, const QString& backupFilePath, QString* error)
 {
     Q_ASSERT(!m_data.filePath.isEmpty());
     if (m_data.filePath.isEmpty()) {
@@ -185,7 +214,7 @@ bool Database::save(QString* error, bool atomic, bool backup)
         return false;
     }
 
-    return saveAs(m_data.filePath, error, atomic, backup);
+    return saveAs(m_data.filePath, action, backupFilePath, error);
 }
 
 /**
@@ -203,10 +232,11 @@ bool Database::save(QString* error, bool atomic, bool backup)
  * @param filePath Absolute path of the file to save
  * @param error error message in case of failure
  * @param atomic Use atomic file transactions
- * @param backup Backup the existing database file, if exists
+ * @param backupFilePath Absolute path to the location where the backup should be stored. Passing an empty string
+ * disables backup.
  * @return true on success
  */
-bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool backup)
+bool Database::saveAs(const QString& filePath, SaveAction action, const QString& backupFilePath, QString* error)
 {
     // Disallow overlapping save operations
     if (isSaving()) {
@@ -224,19 +254,7 @@ bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool
         return false;
     }
 
-    // Prevent destructive operations while saving
-    QMutexLocker locker(&m_saveMutex);
-
     if (filePath == m_data.filePath) {
-        // Disallow saving to the same file if read-only
-        if (m_data.isReadOnly) {
-            Q_ASSERT_X(false, "Database::saveAs", "Could not save, database file is read-only.");
-            if (error) {
-                *error = tr("Could not save, database file is read-only.");
-            }
-            return false;
-        }
-
         // Fail-safe check to make sure we don't overwrite underlying file changes
         // that have not yet triggered a file reload/merge operation.
         if (!m_fileWatcher->hasSameFileChecksum()) {
@@ -248,16 +266,22 @@ bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool
     }
 
     // Clear read-only flag
-    setReadOnly(false);
     m_fileWatcher->stop();
 
+    // Add random data to prevent side-channel data deduplication attacks
+    int length = Random::instance()->randomUIntRange(64, 512);
+    m_metadata->customData()->set("KPXC_RANDOM_SLUG", Random::instance()->randomArray(length).toHex());
+
+    // Prevent destructive operations while saving
+    QMutexLocker locker(&m_saveMutex);
+
     QFileInfo fileInfo(filePath);
     auto realFilePath = fileInfo.exists() ? fileInfo.canonicalFilePath() : fileInfo.absoluteFilePath();
     bool isNewFile = !QFile::exists(realFilePath);
-    bool ok = AsyncTask::runAndWaitForFuture([&] { return performSave(realFilePath, error, atomic, backup); });
+    bool ok = AsyncTask::runAndWaitForFuture([&] { return performSave(realFilePath, action, backupFilePath, error); });
     if (ok) {
-        markAsClean();
         setFilePath(filePath);
+        markAsClean();
         if (isNewFile) {
             QFile::setPermissions(realFilePath, QFile::ReadUser | QFile::WriteUser);
         }
@@ -270,9 +294,19 @@ bool Database::saveAs(const QString& filePath, QString* error, bool atomic, bool
     return ok;
 }
 
-bool Database::performSave(const QString& filePath, QString* error, bool atomic, bool backup)
+bool Database::performSave(const QString& filePath, SaveAction action, const QString& backupFilePath, QString* error)
 {
-    if (atomic) {
+    if (!backupFilePath.isNull()) {
+        backupDatabase(filePath, backupFilePath);
+    }
+
+#if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
+    QFileInfo info(filePath);
+    auto createTime = info.exists() ? info.birthTime() : QDateTime::currentDateTime();
+#endif
+
+    switch (action) {
+    case Atomic: {
         QSaveFile saveFile(filePath);
         if (saveFile.open(QIODevice::WriteOnly)) {
             // write the database to the file
@@ -280,9 +314,10 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
                 return false;
             }
 
-            if (backup) {
-                backupDatabase(filePath);
-            }
+#if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
+            // Retain original creation time
+            saveFile.setFileTime(createTime, QFile::FileBirthTime);
+#endif
 
             if (saveFile.commit()) {
                 // successfully saved database file
@@ -293,20 +328,17 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
         if (error) {
             *error = saveFile.errorString();
         }
-    } else {
+        break;
+    }
+    case TempFile: {
         QTemporaryFile tempFile;
         if (tempFile.open()) {
             // write the database to the file
             if (!writeDatabase(&tempFile, error)) {
                 return false;
             }
-
             tempFile.close(); // flush to disk
 
-            if (backup) {
-                backupDatabase(filePath);
-            }
-
             // Delete the original db and move the temp file in place
             auto perms = QFile::permissions(filePath);
             QFile::remove(filePath);
@@ -318,8 +350,12 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
                 // successfully saved the database
                 tempFile.setAutoRemove(false);
                 QFile::setPermissions(filePath, perms);
+#if QT_VERSION >= QT_VERSION_CHECK(5, 10, 0)
+                // Retain orginal creation time
+                tempFile.setFileTime(createTime, QFile::FileBirthTime);
+#endif
                 return true;
-            } else if (!backup || !restoreDatabase(filePath)) {
+            } else if (backupFilePath.isEmpty() || !restoreDatabase(filePath, backupFilePath)) {
                 // Failed to copy new database in place, and
                 // failed to restore from backup or backups disabled
                 tempFile.setAutoRemove(false);
@@ -333,6 +369,23 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
         if (error) {
             *error = tempFile.errorString();
         }
+        break;
+    }
+    case DirectWrite: {
+        // Open the original database file for direct-write
+        QFile dbFile(filePath);
+        if (dbFile.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
+            if (!writeDatabase(&dbFile, error)) {
+                return false;
+            }
+            dbFile.close();
+            return true;
+        }
+        if (error) {
+            *error = dbFile.errorString();
+        }
+        break;
+    }
     }
 
     // Saving failed
@@ -341,17 +394,9 @@ bool Database::performSave(const QString& filePath, QString* error, bool atomic,
 
 bool Database::writeDatabase(QIODevice* device, QString* error)
 {
-    Q_ASSERT(!m_data.isReadOnly);
-    if (m_data.isReadOnly) {
-        if (error) {
-            *error = tr("File cannot be written as it is opened in read-only mode.");
-        }
-        return false;
-    }
-
     PasswordKey oldTransformedKey;
     if (m_data.key->isEmpty()) {
-        oldTransformedKey.setHash(m_data.transformedDatabaseKey->rawKey());
+        oldTransformedKey.setRawKey(m_data.transformedDatabaseKey->rawKey());
     }
 
     KeePass2Writer writer;
@@ -371,7 +416,7 @@ bool Database::writeDatabase(QIODevice* device, QString* error)
     Q_ASSERT(newKey != oldTransformedKey.rawKey());
     if (newKey.isEmpty() || newKey == oldTransformedKey.rawKey()) {
         if (error) {
-            *error = tr("Key not transformed. This is a bug, please report it to the developers!");
+            *error = tr("Key not transformed. This is a bug, please report it to the developers.");
         }
         return false;
     }
@@ -423,6 +468,7 @@ bool Database::import(const QString& xmlExportPath, QString* error)
 void Database::releaseData()
 {
     // Prevent data release while saving
+    Q_ASSERT(!isSaving());
     QMutexLocker locker(&m_saveMutex);
 
     if (m_modified) {
@@ -431,7 +477,6 @@ void Database::releaseData()
 
     setEmitModified(false);
     m_modified = false;
-    m_modifiedTimer.stop();
 
     s_uuidMap.remove(m_uuid);
     m_uuid = QUuid();
@@ -439,32 +484,38 @@ void Database::releaseData()
     m_data.clear();
     m_metadata->clear();
 
-    setRootGroup(new Group());
+    // Reset and delete the root group
+    auto oldGroup = setRootGroup(new Group());
+    delete oldGroup;
 
     m_fileWatcher->stop();
 
     m_deletedObjects.clear();
     m_commonUsernames.clear();
+    m_tagList.clear();
 }
 
 /**
- * Remove the old backup and replace it with a new one
- * backups are named <filename>.old.<extension>
+ * Remove the old backup and replace it with a new one. Backup name is taken from destinationFilePath.
+ * Non-existing parent directories will be created automatically.
  *
  * @param filePath Path to the file to backup
+ * @param destinationFilePath Path to the backup destination file
  * @return true on success
  */
-bool Database::backupDatabase(const QString& filePath)
+bool Database::backupDatabase(const QString& filePath, const QString& destinationFilePath)
 {
-    static auto re = QRegularExpression("(\\.[^.]+)$");
-
-    auto match = re.match(filePath);
-    auto backupFilePath = filePath;
+    // Ensure that the path to write to actually exists
+    auto parentDirectory = QFileInfo(destinationFilePath).absoluteDir();
+    if (!parentDirectory.exists()) {
+        if (!QDir().mkpath(parentDirectory.absolutePath())) {
+            return false;
+        }
+    }
     auto perms = QFile::permissions(filePath);
-    backupFilePath = backupFilePath.replace(re, "") + ".old" + match.captured(1);
-    QFile::remove(backupFilePath);
-    bool res = QFile::copy(filePath, backupFilePath);
-    QFile::setPermissions(backupFilePath, perms);
+    QFile::remove(destinationFilePath);
+    bool res = QFile::copy(filePath, destinationFilePath);
+    QFile::setPermissions(destinationFilePath, perms);
     return res;
 }
 
@@ -476,32 +527,19 @@ bool Database::backupDatabase(const QString& filePath)
  * @param filePath Path to the file to restore
  * @return true on success
  */
-bool Database::restoreDatabase(const QString& filePath)
+bool Database::restoreDatabase(const QString& filePath, const QString& fromBackupFilePath)
 {
-    static auto re = QRegularExpression("^(.*?)(\\.[^.]+)?$");
-
-    auto match = re.match(filePath);
     auto perms = QFile::permissions(filePath);
-    auto backupFilePath = match.captured(1) + ".old" + match.captured(2);
     // Only try to restore if the backup file actually exists
-    if (QFile::exists(backupFilePath)) {
+    if (QFile::exists(fromBackupFilePath)) {
         QFile::remove(filePath);
-        return QFile::copy(backupFilePath, filePath);
-        QFile::setPermissions(filePath, perms);
+        if (QFile::copy(fromBackupFilePath, filePath)) {
+            return QFile::setPermissions(filePath, perms);
+        }
     }
     return false;
 }
 
-bool Database::isReadOnly() const
-{
-    return m_data.isReadOnly;
-}
-
-void Database::setReadOnly(bool readOnly)
-{
-    m_data.isReadOnly = readOnly;
-}
-
 /**
  * Returns true if the database key exists, has subkeys, and the
  * root group exists
@@ -523,14 +561,12 @@ const Group* Database::rootGroup() const
     return m_rootGroup;
 }
 
-/**
- * Sets group as the root group and takes ownership of it.
- * Warning: Be careful when calling this method as it doesn't
- *          emit any notifications so e.g. models aren't updated.
- *          The caller is responsible for cleaning up the previous
-            root group.
+/* Set the root group of the database and return
+ * the old root group. It is the responsibility
+ * of the calling function to dispose of the old
+ * root group.
  */
-void Database::setRootGroup(Group* group)
+Group* Database::setRootGroup(Group* group)
 {
     Q_ASSERT(group);
 
@@ -538,8 +574,17 @@ void Database::setRootGroup(Group* group)
         emit databaseDiscarded();
     }
 
+    auto oldRoot = m_rootGroup;
     m_rootGroup = group;
     m_rootGroup->setParent(this);
+
+    // Initialize the root group if not done already
+    if (m_rootGroup->uuid().isNull()) {
+        m_rootGroup->setUuid(QUuid::createUuid());
+        m_rootGroup->setName(tr("Passwords", "Root group name"));
+    }
+
+    return oldRoot;
 }
 
 Metadata* Database::metadata()
@@ -641,17 +686,57 @@ void Database::addDeletedObject(const QUuid& uuid)
     addDeletedObject(delObj);
 }
 
-QList<QString> Database::commonUsernames()
+const QStringList& Database::commonUsernames() const
 {
     return m_commonUsernames;
 }
 
+const QStringList& Database::tagList() const
+{
+    return m_tagList;
+}
+
 void Database::updateCommonUsernames(int topN)
 {
     m_commonUsernames.clear();
     m_commonUsernames.append(rootGroup()->usernamesRecursive(topN));
 }
 
+void Database::updateTagList()
+{
+    m_tagList.clear();
+    if (!m_rootGroup) {
+        emit tagListUpdated();
+        return;
+    }
+
+    // Search groups recursively looking for tags
+    // Use a set to prevent adding duplicates
+    QSet<QString> tagSet;
+    for (auto entry : m_rootGroup->entriesRecursive()) {
+        if (!entry->isRecycled()) {
+            for (auto tag : entry->tagList()) {
+                tagSet.insert(tag);
+            }
+        }
+    }
+
+    m_tagList = tagSet.toList();
+    m_tagList.sort();
+    emit tagListUpdated();
+}
+
+void Database::removeTag(const QString& tag)
+{
+    if (!m_rootGroup) {
+        return;
+    }
+
+    for (auto entry : m_rootGroup->entriesRecursive()) {
+        entry->removeTag(tag);
+    }
+}
+
 const QUuid& Database::cipher() const
 {
     return m_data.cipher;
@@ -676,11 +761,11 @@ bool Database::challengeMasterSeed(const QByteArray& masterSeed)
 {
     m_keyError.clear();
     if (m_data.key) {
-        m_data.masterSeed->setHash(masterSeed);
+        m_data.masterSeed->setRawKey(masterSeed);
         QByteArray response;
         bool ok = m_data.key->challenge(masterSeed, response, &m_keyError);
         if (ok && !response.isEmpty()) {
-            m_data.challengeResponseKey->setHash(response);
+            m_data.challengeResponseKey->setRawKey(response);
         } else if (ok && response.isEmpty()) {
             // no CR key present, make sure buffer is empty
             m_data.challengeResponseKey.reset(new PasswordKey);
@@ -718,7 +803,6 @@ bool Database::setKey(const QSharedPointer<const CompositeKey>& key,
                       bool updateTransformSalt,
                       bool transformKey)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     m_keyError.clear();
 
     if (!key) {
@@ -734,7 +818,7 @@ bool Database::setKey(const QSharedPointer<const CompositeKey>& key,
 
     PasswordKey oldTransformedDatabaseKey;
     if (m_data.key && !m_data.key->isEmpty()) {
-        oldTransformedDatabaseKey.setHash(m_data.transformedDatabaseKey->rawKey());
+        oldTransformedDatabaseKey.setRawKey(m_data.transformedDatabaseKey->rawKey());
     }
 
     QByteArray transformedDatabaseKey;
@@ -747,7 +831,7 @@ bool Database::setKey(const QSharedPointer<const CompositeKey>& key,
 
     m_data.key = key;
     if (!transformedDatabaseKey.isEmpty()) {
-        m_data.transformedDatabaseKey->setHash(transformedDatabaseKey);
+        m_data.transformedDatabaseKey->setRawKey(transformedDatabaseKey);
     }
     if (updateChangedTime) {
         m_metadata->setDatabaseKeyChanged(Clock::currentDateTimeUtc());
@@ -777,14 +861,11 @@ const QVariantMap& Database::publicCustomData() const
 
 void Database::setPublicCustomData(const QVariantMap& customData)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     m_data.publicCustomData = customData;
 }
 
 void Database::createRecycleBin()
 {
-    Q_ASSERT(!m_data.isReadOnly);
-
     auto recycleBin = new Group();
     recycleBin->setUuid(QUuid::createUuid());
     recycleBin->setParent(rootGroup());
@@ -798,7 +879,6 @@ void Database::createRecycleBin()
 
 void Database::recycleEntry(Entry* entry)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     if (m_metadata->recycleBinEnabled()) {
         if (!m_metadata->recycleBin()) {
             createRecycleBin();
@@ -811,7 +891,6 @@ void Database::recycleEntry(Entry* entry)
 
 void Database::recycleGroup(Group* group)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     if (m_metadata->recycleBinEnabled()) {
         if (!m_metadata->recycleBin()) {
             createRecycleBin();
@@ -824,7 +903,6 @@ void Database::recycleGroup(Group* group)
 
 void Database::emptyRecycleBin()
 {
-    Q_ASSERT(!m_data.isReadOnly);
     if (m_metadata->recycleBinEnabled() && m_metadata->recycleBin()) {
         // destroying direct entries of the recycle bin
         QList<Entry*> subEntries = m_metadata->recycleBin()->entries();
@@ -839,26 +917,22 @@ void Database::emptyRecycleBin()
     }
 }
 
-void Database::setEmitModified(bool value)
+bool Database::isModified() const
 {
-    if (m_emitModified && !value) {
-        m_modifiedTimer.stop();
-    }
-
-    m_emitModified = value;
+    return m_modified;
 }
 
-bool Database::isModified(bool includeNonDataChanges) const
+bool Database::hasNonDataChanges() const
 {
-    return m_modified || (includeNonDataChanges && m_hasNonDataChange);
+    return m_hasNonDataChange;
 }
 
 void Database::markAsModified()
 {
     m_modified = true;
-    if (m_emitModified && !m_modifiedTimer.isActive()) {
+    if (modifiedSignalEnabled() && !m_modifiedTimer.isActive()) {
         // Small time delay prevents numerous consecutive saves due to repeated signals
-        m_modifiedTimer.start(150);
+        startModifiedTimer();
     }
 }
 
@@ -866,7 +940,7 @@ void Database::markAsClean()
 {
     bool emitSignal = m_modified;
     m_modified = false;
-    m_modifiedTimer.stop();
+    stopModifiedTimer();
     m_hasNonDataChange = false;
     if (emitSignal) {
         emit databaseSaved();
@@ -876,6 +950,7 @@ void Database::markAsClean()
 void Database::markNonDataChange()
 {
     m_hasNonDataChange = true;
+    emit databaseNonDataChanged();
 }
 
 /**
@@ -899,14 +974,12 @@ QSharedPointer<Kdf> Database::kdf() const
 
 void Database::setKdf(QSharedPointer<Kdf> kdf)
 {
-    Q_ASSERT(!m_data.isReadOnly);
     m_data.kdf = std::move(kdf);
+    setFormatVersion(KeePass2Writer::kdbxVersionRequired(this, true, m_data.kdf.isNull()));
 }
 
 bool Database::changeKdf(const QSharedPointer<Kdf>& kdf)
 {
-    Q_ASSERT(!m_data.isReadOnly);
-
     kdf->randomizeSeed();
     QByteArray transformedDatabaseKey;
     if (!m_data.key) {
@@ -917,8 +990,20 @@ bool Database::changeKdf(const QSharedPointer<Kdf>& kdf)
     }
 
     setKdf(kdf);
-    m_data.transformedDatabaseKey->setHash(transformedDatabaseKey);
+    m_data.transformedDatabaseKey->setRawKey(transformedDatabaseKey);
     markAsModified();
 
     return true;
 }
+
+// Prevent warning about QTimer not allowed to be started/stopped from other thread
+void Database::startModifiedTimer()
+{
+    QMetaObject::invokeMethod(&m_modifiedTimer, "start", Q_ARG(int, 150));
+}
+
+// Prevent warning about QTimer not allowed to be started/stopped from other thread
+void Database::stopModifiedTimer()
+{
+    QMetaObject::invokeMethod(&m_modifiedTimer, "stop");
+}
